//============================================================================
//
//   SSSS    tt          lll  lll
//  SS  SS   tt           ll   ll
//  SS     tttttt  eeee   ll   ll   aaaa
//   SSSS    tt   ee  ee  ll   ll      aa
//      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
//  SS  SS   tt   ee      ll   ll  aa  aa
//   SSSS     ttt  eeeee llll llll  aaaaa
//
// Copyright (c) 1995-2019 by Bradford W. Mott, Stephen Anthony
// and the Stella Team
//
// See the file "License.txt" for information on usage and redistribution of
// this file, and for a DISCLAIMER OF ALL WARRANTIES.
//============================================================================

#ifndef TIA_PLAYER
#define TIA_PLAYER

#include "bspf.hxx"
#include "Serializable.hxx"
#include "TIAConstants.hxx"

class TIA;

class Player : public Serializable
{
  public:
    explicit Player(uInt32 collisionMask);

  public:

    void setTIA(TIA* tia) { myTIA = tia; }

    void reset();

    void grp(uInt8 value);

    void hmp(uInt8 value);

    void nusiz(uInt8 value, bool hblank);

    void resp(uInt8 counter);

    void refp(uInt8 value);

    void vdelp(uInt8 value);

    void toggleEnabled(bool enabled);

    void toggleCollisions(bool enabled);

    void setColor(uInt8 color);

    void setDebugColor(uInt8 color);
    void enableDebugColors(bool enabled);

    void applyColorLoss();

    void setInvertedPhaseClock(bool enable);

    void startMovement();

    void nextLine();

    uInt8 getClock() const { return myCounter; }

    bool isOn() const { return (collision & 0x8000); }
    uInt8 getColor() const { return myColor; }

    void shufflePatterns();

    uInt8 getRespClock() const;

    uInt8 getPosition() const;
    void setPosition(uInt8 newPosition);

    uInt8 getGRPOld() const { return myPatternOld; }
    uInt8 getGRPNew() const { return myPatternNew; }

    void setGRPOld(uInt8 pattern);

    /**
      Serializable methods (see that class for more information).
    */
    bool save(Serializer& out) const override;
    bool load(Serializer& in) override;

    void movementTick(uInt32 clock, bool hblank)
    {
      if (clock == myHmmClocks)
        isMoving = false;

      if(isMoving)
      {
        if (hblank) tick();
        myInvertedPhaseClock = !hblank;
      }
    }

    void tick()
    {
      if(myUseInvertedPhaseClock && myInvertedPhaseClock)
      {
        myInvertedPhaseClock = false;
        return;
      }

      if (!myIsRendering || myRenderCounter < myRenderCounterTripPoint)
        collision = myCollisionMaskDisabled;
      else
        collision = (myPattern & (1 << mySampleCounter)) ? myCollisionMaskEnabled : myCollisionMaskDisabled;

      if (myDecodes[myCounter]) {
        myIsRendering = true;
        mySampleCounter = 0;
        myRenderCounter = renderCounterOffset;
      } else if (myIsRendering) {
        ++myRenderCounter;

        switch (myDivider) {
          case 1:
            if (myRenderCounter > 0)
              ++mySampleCounter;

            if (myRenderCounter >= 0 && myDividerChangeCounter >= 0 && myDividerChangeCounter-- == 0)
              setDivider(myDividerPending);

            break;

          default:
            if (myRenderCounter > 1 && (((myRenderCounter - 1) % myDivider) == 0))
              ++mySampleCounter;

            if (myRenderCounter > 0 && myDividerChangeCounter >= 0 && myDividerChangeCounter-- == 0)
              setDivider(myDividerPending);

            break;
        }

        if (mySampleCounter > 7) myIsRendering = false;
      }

      if (++myCounter >= TIAConstants::H_PIXEL) myCounter = 0;
    }

  public:

    uInt32 collision;
    bool isMoving;

  private:

    void updatePattern();
    void applyColors();
    void setDivider(uInt8 divider);

  private:

    enum Count: Int8 {
      renderCounterOffset = -5,
    };

  private:

    uInt32 myCollisionMaskDisabled;
    uInt32 myCollisionMaskEnabled;

    uInt8 myColor;
    uInt8 myObjectColor, myDebugColor;
    bool myDebugEnabled;

    bool myIsSuppressed;

    uInt8 myHmmClocks;
    uInt8 myCounter;

    bool myIsRendering;
    Int8 myRenderCounter;
    Int8 myRenderCounterTripPoint;
    uInt8 myDivider;
    uInt8 myDividerPending;
    uInt8 mySampleCounter;
    Int8 myDividerChangeCounter;

    const uInt8* myDecodes;
    uInt8 myDecodesOffset;  // needed for state saving

    uInt8 myPatternOld;
    uInt8 myPatternNew;
    uInt8 myPattern;

    bool myIsReflected;
    bool myIsDelaying;
    bool myInvertedPhaseClock;
    bool myUseInvertedPhaseClock;

    TIA* myTIA;

  private:
    Player(const Player&) = delete;
    Player(Player&&) = delete;
    Player& operator=(const Player&) = delete;
    Player& operator=(Player&&) = delete;
};

#endif // TIA_PLAYER
